import inspect
import unicodedata

from importlib import import_module

from django.apps import AppConfig
from django.utils.functional import cached_property
from django.utils.html import escape
from django.utils.safestring import SafeData, mark_safe
from django.utils.text import normalize_newlines


class InvalidFunctionArguments(Exception):
    pass


def remove_accent(string_with_diacritics):
    """
    Removes a diacritics from a given string"
    """
    return unicodedata.normalize('NFKD', string_with_diacritics).encode('ASCII', 'ignore').decode('ASCII')


def get_class_method(cls_or_inst, method_name):
    """
    Returns a method from a given class or instance. When the method doest not exist, it returns `None`. Also works with
    properties and cached properties.
    """
    cls = cls_or_inst if isinstance(cls_or_inst, type) else cls_or_inst.__class__
    meth = getattr(cls, method_name, None)
    if isinstance(meth, property):
        meth = meth.fget
    elif isinstance(meth, cached_property):
        meth = meth.func
    return meth if callable(meth) else None


def keep_spacing(value, autoescape=True):
    """
    When a given `str` contains multiple spaces it keeps first space and others are replaces by &nbsp;. Newlines
    converts into HTML's <br>.
    """
    autoescape = autoescape and not isinstance(value, SafeData)
    value = normalize_newlines(value)
    if autoescape:
        value = escape(value)
    return mark_safe(value.replace('  ', ' &nbsp;').replace('\n', '<br />'))


def call_function_with_unknown_input(function, **kwargs):
    """
    Call function and use kwargs from input if function requires them.
    :param function: function to call.
    :param kwargs: function input kwargs or extra kwargs which will not be used.
    :return: function result or raised InvalidFunctionArguments exception.
    """
    function_kwargs_names = set(inspect.signature(function).parameters.keys())

    required_function_kwargs_names = set(
        k for k, v in inspect.signature(function).parameters.items() if v.default is v.empty
    )

    function_kwargs = {k: v for k, v in kwargs.items() if k in function_kwargs_names}
    if required_function_kwargs_names <= set(function_kwargs.keys()):
        return function(**function_kwargs)
    else:
        raise InvalidFunctionArguments(
            'Function {} missing required arguments {}'.format(
                function, ', '.join(required_function_kwargs_names - set(function_kwargs.keys()))
            )
        )


def generate_container_app_config(name,):
    name_prefix, _, app_name = name.split('.')

    label = '{}_{}'.format(name_prefix, app_name)
    cls_name = '{}{}AppConfig'.format(name_prefix.title(), app_name.title())

    cls = type(cls_name, (AppConfig,), {'name': name, 'label': label})

    return cls(name, import_module(cls.name))


ContainerAppConfig = generate_container_app_config



from django.core.exceptions import ImproperlyConfigured

from chamber.utils.transaction import pre_commit, UniquePreCommitCallable
from chamber.models.signals import dispatcher_post_save


class BaseHandler:
    """
    Base dispatcher class that can be subclassed to call a handler based on a change in some a SmartModel.
    If you subclass, be sure the __call__ method does not change signature.
    """

    signal = None

    def __init__(self, signal=None):
        self._signal = signal if signal is not None else self.signal

    def connect(self, sender):
        if not self._signal:
            raise ImproperlyConfigured('Handler cannot be connected as dispatcher, because signal is not set')
        self._signal.connect(self, sender=sender)

    def __call__(self, instance, **kwargs):
        """
        `instance` ... instance of the SmartModel where the handler is being called
        Some dispatchers require additional params to evaluate the handler can be dispatched,
        these are hidden in args and kwargs.
        """
        if self.can_handle(instance, **kwargs):
            self._handle(instance, **kwargs)

    def _handle(self, instance, **kwargs):
        self.handle(instance, **kwargs)

    def can_handle(self, instance, **kwargs):
        return True

    def handle(self, instance, **kwargs):
        raise NotImplementedError


class PreCommitHandler(BaseHandler):
    """
    Handler class that is used for performing on success operations.
    """

    signal = dispatcher_post_save

    def __init__(self, using=None, *args, **kwargs):
        self.using = using
        super().__init__(*args, **kwargs)

    def _handle(self, instance, **kwargs):
        pre_commit(lambda: self.handle(instance, **kwargs), using=self.using)


class InstanceOneTimePreCommitHandlerCallable(UniquePreCommitCallable):
    """
    Use this class to create on success caller that will be unique per instance and will be called only once per
    instance.
    """

    def __init__(self, handler, instance):
        super().__init__(instance=instance)
        self.handler = handler

    def _get_instance(self):
        instance = self.kwargs_list[0]['instance']
        return instance.__class__.objects.get(pk=instance.pk)

    def _get_unique_id(self):
        instance = self.kwargs_list[0]['instance']
        return hash((self.handler.__class__, instance.__class__, instance.pk))

    def __call__(self):
        self.handler.handle(instance=self._get_instance())


class InstanceOneTimePreCommitHandler(PreCommitHandler):
    """
    Use this class to create handler that will be unique per instance and will be called only once per instance.
    """

    def _handle(self, instance, **kwargs):
        pre_commit(InstanceOneTimePreCommitHandlerCallable(self, instance), using=self.using)



from urllib.parse import urlparse

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured


class Domain:

    def __init__(self, name, urlconf=None, user_model=None, url=None, protocol=None, hostname=None, port=None):
        self.name = name
        self.protocol = protocol
        self.hostname = hostname
        self.urlconf = urlconf
        self.port = port

        if url:
            parsed_url = urlparse(url)
            self.protocol, self.hostname, self.port = parsed_url.scheme, parsed_url.hostname, parsed_url.port

        if self.protocol is None:
            raise ImproperlyConfigured('protocol must be set')
        if self.hostname is None:
            raise ImproperlyConfigured('hostname must be set')

        if self.port is None:
            if self.protocol == 'http':
                self.port = 80
            elif self.protocol == 'https':
                self.port = 443
            else:
                raise ImproperlyConfigured('port must be set')
        self.user_model = user_model

    @property
    def url(self):
        return ('{}://{}'.format(self.protocol, self.hostname)
                if (self.protocol == 'http' and self.port == 80) or (self.protocol == 'https' and self.port == 443)
                else '{}://{}:{}'.format(self.protocol, self.hostname, self.port))

    @property
    def user_class(self):
        return apps.get_model(*self.user_model.split('.', 1))


def get_domain(site_id):
    from django.conf import settings

    if site_id in settings.DOMAINS:
        return settings.DOMAINS.get(site_id)
    else:
        raise ImproperlyConfigured('Domain with ID "{}" does not exists'.format(site_id))


def get_current_domain():
    from django.conf import settings

    return get_domain(settings.SITE_ID)


def get_user_class():
    return get_current_domain().user_class


def get_domain_choices():
    from django.conf import settings

    return [(key, domain.name) for key, domain in settings.DOMAINS.items()]



from django.core.serializers.json import DjangoJSONEncoder


class ChamberJSONEncoder(DjangoJSONEncoder):

    def default(self, val):
        try:
            return super().default(val)
        except TypeError:
            # https://github.com/python/cpython/blob/v3.8.3/Lib/json/encoder.py#L160-L180
            return 'Unserializable value of "{}"'.format(str(type(val)))
        except ValueError as ex:
            # https://github.com/django/django/blob/2.2.14/django/core/serializers/json.py#L81-L104
            return str(ex)



import json
import logging
import platform

from django.http import UnreadablePostError

from chamber.utils.json import ChamberJSONEncoder


def skip_unreadable_post(record):
    if record.exc_info:
        exc_type, exc_value = record.exc_info[:2]
        if isinstance(exc_value, UnreadablePostError):
            return False
    return True


class AppendExtraJSONHandler(logging.StreamHandler):

    DEFAULT_STREAM_HANDLER_VARIABLE_KEYS = {
        'name', 'msg', 'args', 'levelname', 'levelno', 'pathname', 'filename', 'module', 'exc_info', 'exc_text',
        'stack_info', 'lineno', 'funcName', 'created', 'msecs', 'relativeCreated', 'thread', 'threadName',
        'processName', 'process',
    }
    CUSTOM_STREAM_HANDLER_VARIABLE_KEYS = {'hostname'}

    def emit(self, record):
        extra = {
            k: v
            for k, v in record.__dict__.items()
            if k not in self.DEFAULT_STREAM_HANDLER_VARIABLE_KEYS.union(self.CUSTOM_STREAM_HANDLER_VARIABLE_KEYS)
        }
        record.msg = '{} --- {}'.format(record.msg, json.dumps(extra, cls=ChamberJSONEncoder))
        super().emit(record)


class HostnameFilter(logging.Filter):

    hostname = platform.node()

    def filter(self, record):
        record.hostname = self.hostname
        return True



from django.db import transaction, models, OperationalError
from django.db.models.manager import BaseManager
from django.db.models.base import ModelBase
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _
from django.utils.functional import cached_property

from chamber.exceptions import PersistenceException
from chamber.patch import Options
from chamber.shortcuts import change_and_save, change, bulk_change_and_save

from .changed_fields import DynamicChangedFields
from .signals import dispatcher_post_save, dispatcher_pre_save


class SmartQuerySetMixin:

    def fast_distinct(self):
        """
        Because standard distinct used on the all fields are very slow and works only with PostgreSQL database
        this method provides alternative to the standard distinct method.
        :return: qs with unique objects
        """
        qs = self.model.objects.filter(pk__in=self.values('pk'))
        if self.query.order_by:
            qs = qs.order_by(*self.query.order_by)
        return qs

    def change_and_save(self, update_only_changed_fields=False, **changed_fields):
        """
        Changes a given `changed_fields` on each object in the queryset, saves objects
        and returns the changed objects in the queryset.
        """
        bulk_change_and_save(self, update_only_changed_fields=update_only_changed_fields, **changed_fields)
        return self.filter()

    def first(self, *field_names):
        """
        Adds possibility to set order fields to default Django first method.
        """
        if field_names:
            return self.order_by(*field_names).first()
        else:
            return super().first()

    def last(self, *field_names):
        """
        Adds possibility to set order fields to default Django last method.
        """
        if field_names:
            return self.order_by(*field_names).last()
        else:
            return super().last()


class SmartQuerySet(SmartQuerySetMixin, models.QuerySet):
    pass


class SmartManager(BaseManager.from_queryset(SmartQuerySet)):
    pass


class SmartModelBase(ModelBase):
    """
    Smart model meta class that register dispatchers to the post or pre save signals.
    """

    def __new__(cls, name, bases, attrs):

        new_cls = super().__new__(cls, name, bases, attrs)
        for dispatcher in new_cls.dispatchers:
            dispatcher.connect(new_cls)
        return new_cls


class AuditModel(models.Model):

    created_at = models.DateTimeField(
        verbose_name=_('created at'),
        null=False,
        blank=False,
        auto_now_add=True,
        db_index=True
    )
    changed_at = models.DateTimeField(
        verbose_name=_('changed at'),
        null=False,
        blank=False,
        auto_now=True,
        db_index=True
    )

    class Meta:
        abstract = True


class SmartModel(AuditModel, metaclass=SmartModelBase):

    objects = SmartManager()

    dispatchers = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_adding = True
        self.is_changing = False
        self._changed_fields = DynamicChangedFields(self)

    class Meta:
        abstract = True

    def __str__(self):
        return '{} #{}'.format(self._meta.verbose_name, self.pk)

    @classmethod
    def from_db(cls, db, field_names, values):
        new = super().from_db(db, field_names, values)
        new.is_adding = False
        new.is_changing = True
        updating_fields = [
            f.name for f in cls._meta.concrete_fields
            if len(values) == len(cls._meta.concrete_fields) or f.attname in field_names
        ]
        new._changed_fields.from_db(fields=updating_fields)
        return new

    @property
    def has_changed(self):
        return bool(self._changed_fields)

    @property
    def changed_fields(self):
        return self._changed_fields

    @property
    def initial_values(self):
        return self._changed_fields.initial_values

    def full_clean(self, exclude=None, *args, **kwargs):
        errors = {}
        for field in self._meta.fields:
            if (not exclude or field.name not in exclude) and hasattr(self, 'clean_{}'.format(field.name)):
                try:
                    getattr(self, 'clean_{}'.format(field.name))()
                except ValidationError as er:
                    errors[field.name] = er

        if errors:
            raise ValidationError(errors)
        super().full_clean(exclude=exclude, *args, **kwargs)

    def _clean_save(self, *args, **kwargs):
        self._persistence_clean(*args, **kwargs)

    def _clean_delete(self, *args, **kwargs):
        self._persistence_clean(*args, **kwargs)

    def _clean_pre_save(self, *args, **kwargs):
        self._clean_save(*args, **kwargs)

    def _clean_pre_delete(self, *args, **kwargs):
        self._clean_delete(*args, **kwargs)

    def _clean_post_save(self, *args, **kwargs):
        self._clean_save(*args, **kwargs)

    def _clean_post_delete(self, *args, **kwargs):
        self._clean_delete(*args, **kwargs)

    def _persistence_clean(self, *args, **kwargs):
        exclude = kwargs.pop('exclude', None)
        try:
            self.full_clean(exclude=exclude)
        except ValidationError as er:
            if hasattr(er, 'error_dict'):
                raise PersistenceException(
                    ', '.join(
                        ('%s: %s' % (key, ', '.join(map(str, val))) for key, val in er.message_dict.items())
                    ),
                    error_dict=er.message_dict
                )
            else:
                raise PersistenceException(', '.join(map(str, er.messages)))

    def _get_save_extra_kwargs(self):
        return {}

    def _pre_save(self, changed, changed_fields, *args, **kwargs):
        """
        :param change: True if model instance was changed, False if was created
        :param changed_fields: fields that was changed before _pre_save was called (changes in the method do not
                affect it)
        """
        pass

    def _call_pre_save(self, changed, changed_fields, *args, **kwargs):
        self._pre_save(changed, changed_fields, *args, **kwargs)

    def _save(self, update_only_changed_fields=False, is_cleaned_pre_save=None, is_cleaned_post_save=None,
              force_insert=False, force_update=False, using=None, update_fields=None, *args, **kwargs):
        """
        Save of SmartModel has the following sequence:
        * pre-save methods are called
        * pre-save validation is invoked (it can be turned off)
        * pre-save signals are called
        * model is saved, model changed fields are reset, is_adding is not False and is_changing has True value
        * post-save methods are called
        * post-save validation is invoked (it is turned off by default)
        * post-save signals are invoked
        """
        is_cleaned_pre_save = (
            self._smart_meta.is_cleaned_pre_save if is_cleaned_pre_save is None else is_cleaned_pre_save
        )
        is_cleaned_post_save = (
            self._smart_meta.is_cleaned_post_save if is_cleaned_post_save is None else is_cleaned_post_save
        )

        origin = self.__class__

        kwargs.update(self._get_save_extra_kwargs())

        self._call_pre_save(
            changed=self.is_changing, changed_fields=self.changed_fields.get_static_changes(), *args, **kwargs
        )
        if is_cleaned_pre_save:
            self._clean_pre_save(*args, **kwargs)
        dispatcher_pre_save.send(
            sender=origin, instance=self, changed=self.is_changing,
            changed_fields=self.changed_fields.get_static_changes(),
            *args, **kwargs
        )

        post_save_changed_fields = self.changed_fields.get_static_changes()

        if not update_fields and update_only_changed_fields:
            update_fields = list(post_save_changed_fields.keys()) + ['changed_at']
            # remove primary key from updating fields
            if self._meta.pk.name in update_fields:
                update_fields.remove(self._meta.pk.name)

        # Changed fields must be cached before save, for post_save and signal purposes
        post_save_is_changing = self.is_changing

        self.save_simple(force_insert=force_insert, force_update=force_update, using=using, update_fields=update_fields)

        self._call_post_save(
            changed=post_save_is_changing, changed_fields=post_save_changed_fields, *args, **kwargs
        )
        if is_cleaned_post_save:
            self._clean_post_save(*args, **kwargs)
        dispatcher_post_save.send(
            sender=origin, instance=self, changed=post_save_is_changing, changed_fields=post_save_changed_fields,
            *args, **kwargs
        )

    def _post_save(self, changed, changed_fields, *args, **kwargs):
        """
        :param change: True if model instance was changed, False if was created
        :param changed_fields: fields that was changed before _post_save was called (changes in the method do not
               affect it)
         """
        pass

    def _call_post_save(self, changed, changed_fields, *args, **kwargs):
        self._post_save(changed, changed_fields, *args, **kwargs)

    def save_simple(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self.is_adding = False
        self.is_changing = True
        self._changed_fields.from_db()

    def save(self, update_only_changed_fields=False, *args, **kwargs):
        if self._smart_meta.is_save_atomic:
            with transaction.atomic():
                self._save(update_only_changed_fields=update_only_changed_fields, *args, **kwargs)
        else:
            self._save(update_only_changed_fields=update_only_changed_fields, *args, **kwargs)

    def _pre_delete(self, *args, **kwargs):
        pass

    def _delete(self, is_cleaned_pre_delete=None, is_cleaned_post_delete=None, *args, **kwargs):
        is_cleaned_pre_delete = (
            self._smart_meta.is_cleaned_pre_delete if is_cleaned_pre_delete is None else is_cleaned_pre_delete
        )
        is_cleaned_post_delete = (
            self._smart_meta.is_cleaned_post_delete if is_cleaned_post_delete is None else is_cleaned_post_delete
        )

        self._pre_delete(*args, **kwargs)

        if is_cleaned_pre_delete:
            self._clean_pre_delete(*args, **kwargs)

        super().delete(*args, **kwargs)

        self._post_delete(*args, **kwargs)

        if is_cleaned_post_delete:
            self._clean_post_delete(*args, **kwargs)

    def _post_delete(self, *args, **kwargs):
        pass

    def delete(self, *args, **kwargs):
        if self._smart_meta.is_delete_atomic:
            with transaction.atomic():
                self._delete(*args, **kwargs)
        else:
            self._delete(*args, **kwargs)

    def refresh_from_db(self, using=None, fields=None):
        super().refresh_from_db(using=using, fields=fields)
        for key, value in self.__class__.__dict__.items():
            if isinstance(value, cached_property):
                self.__dict__.pop(key, None)
        self.is_adding = False
        self.is_changing = True

        self._changed_fields.from_db(fields={
            f.name for f in self._meta.concrete_fields if not fields or f.attname in fields or f.name in fields
        })

        return self

    def change(self, **changed_fields):
        """
        Changes a given `changed_fields` on this object and returns itself.
        :param changed_fields: fields to change
        :return: self
        """
        change(self, **changed_fields)
        return self

    def change_and_save(self, update_only_changed_fields=False, **changed_fields):
        """
        Changes a given `changed_fields` on this object, saves it and returns itself.
        :param update_only_changed_fields: only changed fields will be updated in the database.
        :param changed_fields: fields to change.
        :return: self
        """
        change_and_save(self, update_only_changed_fields=update_only_changed_fields, **changed_fields)
        return self

    def get_locked_instance(self):
        """
        Lock object and reload it from database.
        :return: reloaded locked object from database
        """
        if not self.pk:
            raise OperationalError('Unsaved object cannot be locked')

        return self.__class__.objects.filter(pk=self.pk).select_for_update().get()


class SmartOptions(Options):

    meta_class_name = 'SmartMeta'
    meta_name = '_smart_meta'
    model_class = SmartModel
    attributes = {
        'is_cleaned_pre_save': True,
        'is_cleaned_post_save': False,
        'is_cleaned_pre_delete': False,
        'is_cleaned_post_delete': False,
        'is_save_atomic': False,
        'is_delete_atomic': False,
    }



from datetime import date, datetime, time

from django.http.response import Http404
from django.shortcuts import _get_queryset
from django.utils import timezone
from django.core.exceptions import ValidationError


def get_object_or_none(klass, *args, **kwargs):
    queryset = _get_queryset(klass)
    try:
        return queryset.get(*args, **kwargs)
    except (queryset.model.DoesNotExist, ValueError, ValidationError):
        return None


def get_object_or_404(klass, *args, **kwargs):
    queryset = _get_queryset(klass)
    try:
        return queryset.get(*args, **kwargs)
    except (queryset.model.DoesNotExist, ValueError, ValidationError):
        raise Http404


def distinct_field(klass, *args, **kwargs):
    return _get_queryset(klass).order_by().values_list(*args, **kwargs).distinct()


def filter_or_exclude_by_date(negate, klass, **kwargs):
    filter_kwargs = {}
    for key, date_value in kwargs.items():
        assert isinstance(date_value, date)

        date_range = (
            timezone.make_aware(datetime.combine(date_value, time.min), timezone.get_current_timezone()),
            timezone.make_aware(datetime.combine(date_value, time.max), timezone.get_current_timezone())
        )
        filter_kwargs['%s__range' % key] = date_range

    if negate:
        return _get_queryset(klass).exclude(**filter_kwargs)
    else:
        return _get_queryset(klass).filter(**filter_kwargs)


def filter_by_date(klass, **kwargs):
    return filter_or_exclude_by_date(False, klass, **kwargs)


def exclude_by_date(klass, **kwargs):
    return filter_or_exclude_by_date(True, klass, **kwargs)


def get_model_field_names(model):
    return {
        field.name for field in model._meta.fields
    } | {
        field.attname for field in model._meta.fields
    } | {'pk'}


def change(obj, **changed_fields):
    """
    Changes a given `changed_fields` on object and returns changed object.
    """
    obj_field_names = get_model_field_names(obj)

    for field_name, value in changed_fields.items():
        if field_name not in obj_field_names:
            raise ValueError("'{}' is an invalid field name".format(field_name))
        setattr(obj, field_name, value)
    return obj


def get_update_fields(obj, **changed_fields):
    obj_field_names = get_model_field_names(obj)
    update_fields = []
    for field_name, value in changed_fields.items():
        if field_name not in obj_field_names:
            raise ValueError("'{}' is an invalid field name".format(field_name))

        if getattr(obj, field_name) != value:
            update_fields.append(field_name)
    return update_fields


def change_and_save(obj, update_only_changed_fields=False, save_kwargs=None, **changed_fields):
    """
    Changes a given `changed_fields` on object, saves it and returns changed object.
    """
    from chamber.models import SmartModel

    save_kwargs = save_kwargs if save_kwargs is not None else {}
    if update_only_changed_fields:
        if isinstance(obj, SmartModel):
            save_kwargs['update_only_changed_fields'] = True
        else:
            save_kwargs['update_fields'] = get_update_fields(obj, **changed_fields)

    change(obj, **changed_fields)
    obj.save(**save_kwargs)
    return obj


def bulk_change(iterable, **changed_fields):
    """
    Changes a given `changed_fields` on each object in a given `iterable`, returns the changed objects.
    """
    return [change(obj, **changed_fields) for obj in iterable]


def bulk_change_and_save(iterable, update_only_changed_fields=False, save_kwargs=None, **changed_fields):
    """
    Changes a given `changed_fields` on each object in a given `iterable`, saves objects
    and returns the changed objects.
    """
    return [
        change_and_save(obj, update_only_changed_fields=update_only_changed_fields, save_kwargs=save_kwargs,
                        **changed_fields)
        for obj in iterable
    ]


def bulk_save(iterable):
    """
    Saves a objects in a given `iterable`.
    """
    return [obj.save() for obj in iterable]



from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.core.management import call_command


class Command(BaseCommand):

    def handle(self, **options):
        if not settings.INITAL_DATA_PATH:
            raise CommandError('Missing "INITAL_DATA_PATH" configuration')
        call_command('loaddata', settings.INITAL_DATA_PATH, stdout=self.stdout, stderr=self.stderr)



# Apply patch only if django is installed
try:
    from django.core.exceptions import ImproperlyConfigured
    try:
        from chamber.patch import *  # NOQA
    except ImproperlyConfigured:
        pass
except ImportError:
    pass



from django.conf import settings as django_settings


DEFAULTS = {
    'MAX_FILE_UPLOAD_SIZE': 20,
    'MULTIDOMAINS_OVERTAKER_AUTH_COOKIE_NAME': None,
    'DEFAULT_IMAGE_ALLOWED_CONTENT_TYPES': {'image/jpeg', 'image/png', 'image/gif'},
    'PRIVATE_S3_STORAGE_URL_EXPIRATION': 3600,
    'AWS_S3_ON': getattr(django_settings, 'AWS_S3_ON', False),
    'AWS_REGION': getattr(django_settings, 'AWS_REGION', None),
}


class Settings:

    def __getattr__(self, attr):
        if attr not in DEFAULTS:
            raise AttributeError('Invalid CHAMBER setting: "{}"'.format(attr))

        default = DEFAULTS[attr]
        return getattr(django_settings, 'CHAMBER_{}'.format(attr), default(self) if callable(default) else default)


settings = Settings()



import collections

import copy

from chamber.utils.decorators import singleton


def _should_exclude_field(field_name, fields, exclude):
    return (fields and field_name not in fields) or (exclude and field_name in exclude)


def field_value_from_instance(field, instance):
    """
    Converts a model field to a dictionary
    """
    value = field.value_from_object(instance)
    if isinstance(value, (dict, list)):
        return copy.deepcopy(value)
    else:
        return value


@singleton
class UnknownSingleton:

    def __repr__(self):
        return 'unknown'

    def __bool__(self):
        return False


Unknown = UnknownSingleton()


@singleton
class DeferredSingleton:

    def __repr__(self):
        return 'deferred'

    def __bool__(self):
        return False


Deferred = DeferredSingleton()


def get_model_fields(model):
    return [field for field in model._meta.concrete_fields]  # pylint: disable=W0212


def get_model_field_names(model):
    return [field.name for field in get_model_fields(model)]


def unknown_model_fields_to_dict(instance, fields=None, exclude=None):
    return {
        field_name: Unknown
        for field_name in get_model_field_names(instance)
        if not _should_exclude_field(field_name, fields, exclude)
    }


def model_to_dict(instance, fields=None, exclude=None):
    """
    The same implementation as django model_to_dict but editable fields are allowed
    """
    return {
        field.name: copy.deepcopy(field_value_from_instance(field, instance))
        for field in get_model_fields(instance)  # pylint: disable=W0212
        if not _should_exclude_field(field.name, fields, exclude)
    }


ValueChange = collections.namedtuple('ValueChange', ('initial', 'current'))


class ChangedFields:
    """
    Class stores changed fields and its initial and current values.
    """

    def __init__(self, initial_dict):
        self._initial_dict = initial_dict

    @property
    def initial_values(self):
        return self._initial_dict.copy()

    @property
    def current_values(self):
        return self.get_current_values()

    def get_current_values(self, fields=None):
        raise NotImplementedError

    @property
    def changed_values(self):
        return {k: value_change.current for k, value_change in self.get_diff().items()}

    def get_diff(self, fields=None):
        d1 = self.initial_values
        d2 = self.get_current_values(fields=fields)
        return {k: ValueChange(d1[k], v) for k, v in d2.items() if v != d1[k]}

    def __setitem__(self, key, item):
        raise AttributeError('Object is readonly')

    def __getitem__(self, key):
        return self.get_diff(fields=[key])[key]

    def __bool__(self):
        return bool(self.get_diff())

    def __len__(self):
        return len(self.get_diff())

    def __delitem__(self, key):
        raise AttributeError('Object is readonly')

    def clear(self):
        raise AttributeError('Object is readonly')

    def has_key(self, key):
        return self.has_any_key(key)

    def has_any_key(self, *keys):
        diff = self.get_diff(fields=keys)
        return bool(set(diff.keys()) & set(keys))

    def keys(self):
        return self.get_diff().keys()

    def values(self):
        return self.get_diff().values()

    def items(self):
        return self.get_diff().items()

    def pop(self, *args, **kwargs):
        raise AttributeError('Object is readonly')

    def __cmp__(self, dictionary):
        return self.get_diff() == dictionary

    def __contains__(self, item):
        return self.has_any_key(item)

    def __iter__(self):
        return iter(self.get_diff())

    def __repr__(self):
        return repr(self.get_diff())

    def __str__(self):
        return repr(self.get_diff())


class DynamicChangedFields(ChangedFields):
    """
    Dynamic changed fields are changed with the instance changes.
    """

    def __init__(self, instance):
        super().__init__(
            self._get_unknown_dict(instance)
        )
        self.instance = instance

    def _get_unknown_dict(self, instance):
        return unknown_model_fields_to_dict(instance)

    def get_current_values(self, fields=None):
        deferred_values = {
            field_name: value for field_name, value in self._initial_dict.items()
            if field_name in self.instance.get_deferred_fields()
        }
        current_values = model_to_dict(
            self.instance,
            fields=fields,
            exclude=set(deferred_values.keys())
        )
        current_values.update(deferred_values)
        return current_values

    def get_static_changes(self):
        return StaticChangedFields(self.initial_values, self.current_values)

    def from_db(self, fields=None):
        if fields is None:
            fields = {field_name for field_name, value in self._initial_dict.items() if value is not Deferred}

        self._initial_dict.update(
            model_to_dict(self.instance, fields=set(fields))
        )

        for field_name, value in self._initial_dict.items():
            if value is Unknown:
                self._initial_dict[field_name] = Deferred


class StaticChangedFields(ChangedFields):
    """
    Static changed fields are immutable. The origin instance changes will not have an affect.
    """

    def __init__(self, initial_dict, current_dict):
        super().__init__(initial_dict)
        self._current_dict = current_dict

    def get_current_values(self, fields=None):
        return {k: v for k, v in self._current_dict.items() if fields is None or k in fields}



import os

from django.core.management.commands import makemessages


def _remove_pot_creation_date(pofile):
    if os.path.exists(pofile):
        with open(pofile, "rb+") as f:
            lines = f.readlines()
            f.seek(0)
            for line in lines:
                if not line.startswith(b'"POT-Creation-Date: '):
                    f.write(line)
            f.truncate()


class Command(makemessages.Command):
    extra_args = [
        '--keyword=_l',
        '--keyword=_n:1,2',
        '--keyword=_nl:1,2',
        '--keyword=_p:1c,2',
        '--keyword=_np:1c,2,3',
        '--keyword=_pl:1c,2',
        '--keyword=_npl:1c,2,3',
    ]

    xgettext_options = makemessages.Command.xgettext_options + extra_args

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            '--no-creation-date',  action='store_true', dest='no_creation_date', default=False,
            help='Remove POT-Creation-Date from result file.',
        )

    def handle(self, *args, **options):
        self.no_creation_date = options['no_creation_date']
        super().handle(*args, **options)

    def write_po_file(self, potfile, locale):
        super().write_po_file(potfile, locale)
        if self.no_creation_date:
            basedir = os.path.join(os.path.dirname(potfile), locale, 'LC_MESSAGES')
            _remove_pot_creation_date(os.path.join(basedir, '{}.po'.format(self.domain)))



class ComparableModelMixin:

    def equals(self, obj, comparator):
        """
        Use comparator for evaluating if objects are the same
        """
        return comparator.compare(self, obj)


class Comparator:

    def compare(self, a, b):
        """
        Return True if objects are same otherwise False
        """
        raise NotImplementedError



import mimetypes

import magic  # pylint: disable=E0401

from django.core.exceptions import ValidationError
from django.template.defaultfilters import filesizeformat
from django.utils.translation import ugettext


class RestrictedFileValidator:

    def __init__(self, max_upload_size):
        self.max_upload_size = max_upload_size

    def __call__(self, data):
        if data.size > self.max_upload_size:
            raise ValidationError(
                ugettext('Please keep filesize under {max}. Current filesize {current}').format(
                    max=filesizeformat(self.max_upload_size),
                    current=filesizeformat(data.size)
                )
            )
        else:
            return data


class AllowedContentTypesByFilenameFileValidator:

    def __init__(self, content_types):
        self.content_types = content_types

    def __call__(self, data):
        extension_mime_type = mimetypes.guess_type(data.name)[0]

        if extension_mime_type not in self.content_types:
            raise ValidationError(ugettext('Extension of file name is not allowed'))

        return data


class AllowedContentTypesByContentFileValidator:

    def __init__(self, content_types):
        self.content_types = content_types

    def __call__(self, data):
        data.open()
        mime_type = magic.from_buffer(data.read(2048), mime=True)
        data.seek(0)
        if mime_type not in self.content_types:
            raise ValidationError(ugettext('File content was evaluated as not supported file type'))

        return data



from django.utils.translation import ugettext

from chamber.formatters import natural_number_with_currency


def price_humanized(value, inst, currency=None):
    """
    Return a humanized price
    """
    return (natural_number_with_currency(value, ugettext('CZK') if currency is None else currency) if value is not None
            else ugettext('(None)'))



from datetime import datetime, timedelta

from django.core.cache import cache
from django.utils.timezone import localtime


class BatchCachedQuerysetIterator:
    """
    Batch iterator that stores last iterated object in the cache.
    The iterator may be used in different processes (running at a different time).
    Iteration will continue from the last complete batch.

    Eq
        iterator = BatchCachedQuerysetIterator(
            User.objects.all(),  # iterated queryset
            'users',             # key used to store last object (cursor) in cache
            batch_size=100,      # size of one iteration
            expiration=10        # expiration of cached cursor in seconds
        )
        # 100 users are returned (if there is more than 100 users)
        list(iterator)

        iterator2 = BatchCachedQuerysetIterator(User.objects.all(), 'users', batch_size=100, expiration=10)
        # next 100 users are returned (users will be sorted by ID)
        list(iterator2)
    """

    def __init__(self, queryset, key, store_cursor_with_exception=False, batch_size=10000, expiration=None):
        """
        Init batch cached queryset iterator
        :param queryset: queryset which you should want to iterate. Queryset should be sortable by ID
        :param key: unique key of your iterator which will be used for cursor caching
        :param store_cursor_with_exception: store cursor in cache if exception will be occurred
        :param batch_size: size of one batch
        :param expiration: expiration in datatime, timedelta or integer (number of seconds) format
        """
        self._batch_size = batch_size
        self._cache_key = f'batch_queryset_iterator_{key}'
        self._count_processed = 0
        self._queryset = queryset.order_by('pk')
        self._cursor = self._cached_cursor = self._get_cursor()
        self._chunked_queryset = self._get_chunked_queryset(self._cursor)
        self._expiration = self._compute_expiration(expiration)
        self._store_cursor_with_exception = store_cursor_with_exception

    def _compute_expiration(self, expiration):
        if isinstance(expiration, datetime):
            return expiration
        elif isinstance(expiration, int):
            return localtime() + timedelta(seconds=expiration)
        elif isinstance(expiration, timedelta):
            return localtime() + expiration
        else:
            raise AttributeError('invalid value of expiration it can be datetime, integer or timedelta')

    def _get_cursor(self):
        return cache.get(self._cache_key)

    def _set_cursor(self):
        if self._cursor != self._cached_cursor:
            cache.set(self._cache_key, self._cursor, (self._expiration - localtime()).total_seconds())
            self._cached_cursor = self._cursor

    def _filter_queryset_by_cursor(self, cursor):
        queryset = self._queryset
        if cursor:
            queryset = queryset.filter(pk__gt=cursor)
        return queryset

    def _get_chunked_queryset(self, cursor):
        return self._filter_queryset_by_cursor(cursor)[:self._batch_size]

    def __iter__(self):
        try:
            for obj in self._get_chunked_queryset(self._cursor):
                yield obj
                self._cursor = obj.pk
            self._set_cursor()
        finally:
            if self._store_cursor_with_exception:
                self._set_cursor()

    def __len__(self):
        return self._get_chunked_queryset(self._cursor).count()

    @property
    def total_number_of_objects(self):
        return self._queryset.count()

    @property
    def remaining_number_of_objects(self):
        return self._filter_queryset_by_cursor(cursor=self._cursor).count()



import logging

from django.core.exceptions import PermissionDenied
from django.http import HttpResponse
from django.utils.functional import SimpleLazyObject

from auth_token import utils  # pylint: disable=E0401
from auth_token.config import settings as auth_token_settings  # pylint: disable=E0401
from auth_token.middleware import TokenAuthenticationMiddleware, get_user  # pylint: disable=E0401
from auth_token.models import AuthorizationToken  # pylint: disable=E0401
from auth_token.utils import dont_enforce_csrf_checks, header_name_to_django  # pylint: disable=E0401

from chamber.config import settings
from chamber.shortcuts import get_object_or_none


logger = logging.getLogger(__name__)


def get_token(request):
    """
    Returns the token model instance associated with the given request token key.
    If no user is retrieved AnonymousToken is returned.
    """
    if (not request.META.get(header_name_to_django(auth_token_settings.HEADER_NAME))
            and settings.MULTIDOMAINS_OVERTAKER_AUTH_COOKIE_NAME):
        ovetaker_auth_token = request.COOKIES.get(settings.MULTIDOMAINS_OVERTAKER_AUTH_COOKIE_NAME)
        token = get_object_or_none(AuthorizationToken, key=ovetaker_auth_token, is_active=True)
        if utils.get_user_from_token(token).is_authenticated:
            return token

    return utils.get_token(request)


class MultiDomainsTokenAuthenticationMiddleware(TokenAuthenticationMiddleware):

    def process_request(self, request):
        """
        Lazy set user and token
        """
        request.token = get_token(request)
        request.user = SimpleLazyObject(lambda: get_user(request))
        request._dont_enforce_csrf_checks = dont_enforce_csrf_checks(request)  # pylint: disable=W0212

    def __call__(self, request):
        try:
            response = super().__call__(request)
        except PermissionDenied as exp:
            logger.warning(str(exp))
            response = HttpResponse("Unauthorized", status=401)

        return response



from setuptools import find_packages, setup

from chamber.version import get_version


setup(
    python_requires=">=3.6",
    name='skip-django-chamber',
    version=get_version(),
    description='Utilities library meant as a complement to django-is-core.',
    author='Lubos Matl, Oskar Hollmann',
    author_email='matllubos@gmail.com, oskar@hollmann.me',
    url='http://github.com/skip-pay/django-chamber',
    packages=find_packages(include=['chamber']),
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Framework :: Django',
    ],
    install_requires=[
        'django>=2.2, <4.0',
        'Unidecode>=1.1.1',
        'pyprind>=2.11.2',
        'python-magic>=0.4.27'
    ],
    extras_require={
        'boto3storage': ['django-storages<2.0', 'boto3'],
    },
)



from django.core.management.base import OutputWrapper

from tqdm import tqdm as original_tqdm


class CommandOutputTMDQWrapper:
    """
    Wrapper around stdout/stderr
    """

    def __init__(self, out):
        self._out = out

    def __getattr__(self, name):
        return getattr(self._out, name)

    def isatty(self):
        return hasattr(self._out, 'isatty') and self._out.isatty()

    def write(self, msg):
        self._out.write(msg, ending='')


class tqdm(original_tqdm):

    monitor_interval = 0

    def __init__(self, *args, **kwargs):
        file = kwargs.pop('file', None)
        if file and isinstance(file, OutputWrapper):
            file = CommandOutputTMDQWrapper(file)
        super().__init__(
            *args,
            file=file,
            ncols=100,
            **kwargs,
        )



import logging

from django.db import transaction, DEFAULT_DB_ALIAS
from django.db.transaction import get_connection
from django.db.transaction import TransactionManagementError

from contextlib import contextmanager


logger = logging.getLogger(__name__)


def pre_commit(func, using=None):
    connection = get_connection(using)

    if connection.in_atomic_block:
        # Transaction in progress; save for execution on commit.
        func_hash = hash(func) if isinstance(func, UniquePreCommitCallable) else None
        if (func_hash is None
                or func_hash not in {func_hash for sids, func_hash, func in connection.run_pre_commit}):
            connection.run_pre_commit.append((set(connection.savepoint_ids), func_hash, func))

    elif not connection.get_autocommit():
        raise TransactionManagementError('pre_commit() cannot be used in manual transaction management')
    else:
        # No transaction in progress and in autocommit mode; execute immediately.
        func()


def smart_atomic(using=None, savepoint=True, ignore_errors=None, reversion=True, reversion_using=None):
    """
    Decorator and context manager that overrides django atomic decorator and automatically adds create revision.
    The _atomic closure is required to achieve save ContextDecorator that nest more inner context decorator.
    More here https://stackoverflow.com/questions/45589718/combine-two-context-managers-into-one
    """

    ignore_errors = () if ignore_errors is None else ignore_errors

    @contextmanager
    def _atomic(using=None, savepoint=True, reversion=True, reversion_using=None):
        try:
            from reversion.revisions import create_revision
        except ImportError:
            reversion = False

        if not reversion:
            @contextmanager  # noqa: F811
            def create_revision(*args, **kwargs):  # noqa: F811
                yield

        error = None
        with transaction.atomic(using, savepoint), create_revision(using=reversion_using):
            try:
                yield
            except ignore_errors as ex:
                error = ex

        if error:
            raise error  # pylint: disable=E0702

    if callable(using):
        return _atomic(DEFAULT_DB_ALIAS, savepoint, reversion, reversion_using)(using)
    else:
        return _atomic(using, savepoint, reversion, reversion_using)


def in_atomic_block(using=None):
    """Check if connection is in atomic block"""

    connection = get_connection(using)
    return connection.in_atomic_block


class UniquePreCommitCallable:
    """
    One time callable class that is used for performing on success operations.
    Handler is callable only once, but data of all calls are stored inside list (kwargs_list).
    """

    def __init__(self, **kwargs):
        self.kwargs_list = (kwargs,)

    def join(self, callable):
        """
        Joins two unique callable.
        """
        self.kwargs_list += callable.kwargs_list

    def _get_unique_id(self):
        """
        Callable instance hash is generated from class name and the return value of this method.
        The method returns None by default, therefore the class init data doesn't have impact.
        You should implement this method to distinguish callable according to its
        input data (for example Django model instance)
        :return:
        """
        return None

    def __hash__(self):
        return hash((self.__class__, self._get_unique_id()))

    def _get_kwargs(self):
        return self.kwargs_list[-1]

    def __call__(self):
        self.handle()

    def handle(self):
        raise NotImplementedError



from urllib.parse import urlencode

from django.conf import settings
from django.urls import reverse as django_reverse


def reverse(viewname, site_id=None, add_domain=False, urlconf=None, args=None, kwargs=None, current_app=None,
            qs_kwargs=None):
    from .domain import get_domain

    urlconf = (get_domain(site_id).urlconf
               if site_id is not None and urlconf is None and settings.SITE_ID != site_id
               else urlconf)
    site_id = settings.SITE_ID if site_id is None else site_id
    domain = get_domain(site_id).url if add_domain else ''
    qs = '?{}'.format(urlencode(qs_kwargs)) if qs_kwargs else ''
    return ''.join((domain, django_reverse(viewname, urlconf, args, kwargs, current_app), qs))



from functools import wraps

from django.conf import settings
from django.utils import translation


class classproperty(property):

    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()


def singleton(klass):
    """
    Create singleton from class
    """
    instances = {}

    def getinstance(*args, **kwargs):
        if klass not in instances:
            instances[klass] = klass(*args, **kwargs)
        return instances[klass]
    return wraps(klass)(getinstance)


def translation_activate_block(function=None, language=None):
    """
    Activate language only for one method or function
    """

    def _translation_activate_block(function):
        def _decorator(*args, **kwargs):
            tmp_language = translation.get_language()
            try:
                translation.activate(language or settings.LANGUAGE_CODE)
                return function(*args, **kwargs)
            finally:
                translation.activate(tmp_language)
        return wraps(function)(_decorator)

    if function:
        return _translation_activate_block(function)
    else:
        return _translation_activate_block



class PersistenceException(Exception):

    def __init__(self, message=None, error_dict=None):
        super().__init__()
        self.message = message
        self.error_dict = error_dict

    def __str__(self):
        return repr(self.message)



import boto3
from botocore.client import Config
from botocore.exceptions import ClientError

from django.core.files.base import ContentFile
from django.core.files.storage import DefaultStorage

from storages.backends.s3boto3 import S3Boto3Storage

from chamber.config import settings


__all__ = (
    'BaseS3Storage',
    'BasePrivateS3Storage',
    'BasePrivateS3DataStorage',
    'force_bytes_content',
    'get_storage_instance',
)


def force_bytes_content(content, blocksize=1024):
    """Returns a tuple of content (file-like object) and bool indicating wheter the content has been casted or not"""
    block = content.read(blocksize)
    content.seek(0)

    if not isinstance(block, bytes):
        _content = bytes(
            content.read(),
            'utf-8' if not hasattr(content, 'encoding') or content.encoding is None else content.encoding,
        )
        return ContentFile(_content), True
    return content, False


def get_storage_class(s3_storage_class, default_storage_class=DefaultStorage):
    return s3_storage_class if settings.AWS_S3_ON else default_storage_class


def get_storage_instance(s3_storage_class, default_storage_class=DefaultStorage):
    return s3_storage_class() if settings.AWS_S3_ON else default_storage_class()


class BaseS3Storage(S3Boto3Storage):

    def _clean_name(self, name):
        # pathlib support
        return super()._clean_name(str(name))

    def _open(self, name, mode='rb'):
        try:
            return super()._open(name, mode)
        except ClientError as ex:
            if ex.response['Error']['Code'] == '403':
                raise PermissionError(f'Cannot open file "{name}": {ex.response["Error"]["Message"]}')
            else:
                raise ex

    def save(self, name, content, max_length=None):
        content, _ = force_bytes_content(content)
        return super().save(name, content, max_length)


class BasePrivateS3Storage(BaseS3Storage):

    expiration = settings.PRIVATE_S3_STORAGE_URL_EXPIRATION

    def url(self, name):
        s3 = boto3.client('s3', config=Config(region_name=settings.AWS_REGION, signature_version='s3v4'))
        url = s3.generate_presigned_url(
            'get_object',
            Params={
                'Bucket': self.bucket_name,
                'Key': self._normalize_name(name),
            },
            ExpiresIn=self.expiration,
        )
        return url


class BasePrivateS3DataStorage(BaseS3Storage):

    def url(self, name):
        raise RuntimeError('You cannot generate data storage URL')



from .base import AuditModel, SmartQuerySet, SmartManager, SmartModel, SmartModelBase   # noqa: F401
from .fields import (   # noqa: F401
    DecimalField, FileField, ImageField, PrevValuePositiveIntegerField, SubchoicesPositiveIntegerField,
    EnumSequencePositiveIntegerField, EnumSequenceCharField, PriceField, PositivePriceField
)
